/* The copyright in this software is being made available under the BSD
 * License, included below. This software may be subject to other third party
 * and contributor rights, including patent rights, and no such rights are
 * granted under this license.  
 *
 * Copyright (c) 2010-2011, ITU/ISO/IEC
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *  * Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *  * Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *  * Neither the name of the ITU/ISO/IEC nor the names of its contributors may
 *    be used to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 */

/** \file     ContextModel.cpp
    \brief    context model class
*/

#include <algorithm>

#include "ContextModel.h"

using namespace std;

// ====================================================================================================================
// Public member functions
// ====================================================================================================================

/**
 - initialize context model with respect to QP and initial probability
 .
 \param  iQp         input QP value
 \param  asCtxInit   initial probability table
 */
Void 
ContextModel::init( Int iQp, Short asCtxInit[] )
{
  Int iInitState  = ( ( asCtxInit[ 0 ] * iQp ) >> 4 ) + asCtxInit[ 1 ];
  iInitState      = min( max( 1, iInitState ), 126 );
  if( iInitState >= 64 )
  {
    m_ucState     = min( 62, iInitState - 64 );
#ifndef PACoder_Prob_Est
    m_ucState    += m_ucState + 1;
#else
		m_ucState = (m_aucState2logprobMPS[m_ucState] << 1) + 1;
#endif
  }
  else
  {
    m_ucState     = min( 62, 63 - iInitState );
#ifndef PACoder_Prob_Est
    m_ucState    += m_ucState;
#else
		m_ucState = (m_aucState2logprobMPS[m_ucState] << 1) + 0;
#endif
  }
}

#ifndef PACoder_Prob_Est
const UChar 
ContextModel::m_aucNextStateMPS[ 64 ] =
{
  1, 2, 3, 4, 5, 6, 7, 8,   
  9, 10,11,12,13,14,15,16,  
  17,18,19,20,21,22,23,24,
  25,26,27,28,29,30,31,32,  
  33,34,35,36,37,38,39,40,  
  41,42,43,44,45,46,47,48,
  49,50,51,52,53,54,55,56,  
  57,58,59,60,61,62,62,63
};

const UChar 
ContextModel::m_aucNextStateLPS[ 64 ] =
{
  0, 0, 1, 2, 2, 4, 4, 5,   
  6, 7, 8, 9, 9, 11,11,12,  
  13,13,15,15,16,16,18,18,
  19,19,21,21,22,22,23,24,  
  24,25,26,26,27,27,28,29,  
  29,30,30,30,31,32,32,33,
  33,33,34,34,35,35,35,36,  
  36,36,37,37,37,38,38,63
};
#else
const UChar ContextModel::m_aucNextStateMPS[ 128 ] = {
  6,   6,   6,   6,   6,   8,  10,  12,  14,  16,  18,  20,  22,  24,  26,  28,
 30,  32,  34,  36,  38,  40,  42,  44,  46,  48,  50,  52,  54,  56,  58,  58,
 60,  64,  66,  66,  70,  70,  72,  76,  78,  78,  80,  82,  84,  88,  90,  90,
 92,  94,  94, 100, 100, 102, 102, 102, 106, 112, 112, 116, 116, 116, 120, 122,
122, 128, 128, 128, 132, 132, 134, 138, 138, 138, 144, 144, 148, 150, 150, 152,
152, 152, 162, 162, 162, 162, 166, 170, 170, 174, 176, 176, 176, 182, 182, 184,
184, 188, 188, 188, 190, 196, 196, 196, 200, 200, 200, 202, 210, 212, 212, 212,
214, 214, 220, 222, 222, 224, 226, 226, 228, 228, 232, 232, 232, 232, 232, 232};

const UChar ContextModel::m_aucNextStateLPS[ 128 ] = {
 30,  30,  30,  30,  40,  40,  40,  40,  42,  42,  44,  44,  44,  46,  46,  52,
 52,  54,  54,  56,  56,  56,  58,  60,  64,  66,  66,  70,  72,  76,  76,  76,
 78,  80,  84,  84,  84,  88,  88,  90,  94, 100, 100, 102, 104, 104, 108, 110,
110, 110, 112, 116, 116, 120, 120, 120, 128, 128, 128, 128, 132, 136, 136, 136,
136, 140, 144, 148, 150, 150, 150, 150, 152, 156, 160, 160, 164, 164, 164, 168,
170, 176, 176, 176, 176, 182, 182, 186, 186, 188, 188, 190, 200, 200, 200, 206,
206, 206, 212, 212, 212, 214, 216, 218, 218, 220, 222, 226, 226, 228, 228, 242,
242, 242, 242, 242, 244, 250, 252, 252, 252, 254, 255, 253, 251, 249, 247, 245};

const UChar ContextModel::m_aucState2logprobMPS[64] = {
127, 116, 110, 106,  95,  91,  82,  76,  72,  66,  66,  61,  55,  55,  50,  47,
 45,  42,  38,  38,  36,  33,  32,  28,  28,  27,  25,  22,  22,  22,  19,  19,
 18,  15,  15,  15,  15,  13,  13,  11,  11,  11,  10,   9,   9,   8,   8,   7,
  7,   6,   6,   6,   6,   5,   5,   5,   4,   4,   4,   3,   3,   3,   3,   3};
#endif

