/* The copyright in this software is being made available under the BSD
 * License, included below. This software may be subject to other third party
 * and contributor rights, including patent rights, and no such rights are
 * granted under this license.
 *
 * Copyright (c) 2010-2011, ITU/ISO/IEC
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *  * Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *  * Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *  * Neither the name of the ITU/ISO/IEC nor the names of its contributors may
 *    be used to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 */

/** \file     TDecBinPACoder.cpp
    \brief    binary entropy decoder of PA-Coder
*/

#include "TDecBinPACoder.h"


TDecBinPACoder::TDecBinPACoder()
: m_pcTComBitstream( 0 )
{
}


TDecBinPACoder::~TDecBinPACoder()
{
}


Void
TDecBinPACoder::init( TComInputBitstream* pcTComBitstream )
{
  m_pcTComBitstream = pcTComBitstream;
}


Void
TDecBinPACoder::uninit()
{
  m_pcTComBitstream = 0;
}


Void
TDecBinPACoder::start()
{
	logprob = 0;
	value = iLastDWORD = m_pcTComBitstream->peekBits(32);
//printf("pacoder mode!\n");
	mapVal2LngPrbNoRenorm(value, logprob, logprbOfVal)
}


Void
TDecBinPACoder::finish()
{
}


Void
TDecBinPACoder::decodeBin( UInt& ruiBin, ContextModel &rcCtxModel )
{
#ifndef PACoder_Prob_Est
	UInt logprobMPS = TComPACoderTables::lgprobMLpsNoRenorm[rcCtxModel.getState()][1];
#else
	UInt logprobMPS = rcCtxModel.getState()+1;
#endif

	if( logprbOfVal > (logprobMPS+logprob) )//fast channel for MPS
	{
		logprob += logprobMPS;
		ruiBin      = rcCtxModel.getMps();
		rcCtxModel.updateMPS();
	}
	else
	{
		UInt thrshld = dvsn_intervalNoRenorm(logprobMPS+logprob);
		if(thrshld <= value)
		{
			value -= thrshld;
#ifndef PACoder_Prob_Est
			logprob += TComPACoderTables::lgprobMLpsNoRenorm[rcCtxModel.getState()][0];
#else
			logprob += TComPACoderTables::ilogprobMPS2logprobLPS[ logprobMPS-1 ];
#endif
			ruiBin      = 1 - rcCtxModel.getMps();
			rcCtxModel.updateLPS();

			mapVal2LngPrbNoRenorm(value, logprob, logprbOfVal)
		}
		else
		{
			logprob += logprobMPS;
		  ruiBin      = rcCtxModel.getMps();
			rcCtxModel.updateMPS();
		}
	}
	
	if(logprob >= 0x800)
	{
		logprob -= 0x800;
		
		m_pcTComBitstream->read(16);
		iLastDWORD = m_pcTComBitstream->peekBits(32);
		value = (value<<16) + (iLastDWORD&0xffff);

		mapVal2LngPrbNoRenorm(value, logprob, logprbOfVal)
	}
}

#ifdef PACoder_LowBound0
Void  TDecBinPACoder::decodeBinL0( UInt& ruiBin, ContextModel& rcCtxModel )
{
	#ifndef PACoder_Prob_Est
	UInt logprobMPS = TComPACoderTables::lgprobMLpsNoRenorm[rcCtxModel.getState()][1];
#else
	UInt logprobMPS = rcCtxModel.getState()+1;
#endif

	if(rcCtxModel.getMps()==0)
	{
		if( logprbOfVal > (logprobMPS+logprob) )//fast channel for MPS
		{
			logprob += logprobMPS;
			ruiBin      = 0;
			rcCtxModel.updateMPS();
		}
		else
		{
			UInt thrshld = dvsn_intervalNoRenorm(logprobMPS+logprob);
			if(thrshld <= value)
			{
				value -= thrshld;
#ifndef PACoder_Prob_Est
				logprob += TComPACoderTables::lgprobMLpsNoRenorm[rcCtxModel.getState()][0];
#else
				logprob += TComPACoderTables::ilogprobMPS2logprobLPS[ logprobMPS-1 ];
#endif
				ruiBin      = 1;
				rcCtxModel.updateLPS();

				mapVal2LngPrbNoRenorm(value, logprob, logprbOfVal)
			}
			else
			{
				logprob += logprobMPS;
				ruiBin      = 0;
				rcCtxModel.updateMPS();
			}
		}
	}
	else
	{
#ifndef PACoder_Prob_Est
		UInt logprobLPS = TComPACoderTables::lgprobMLpsNoRenorm[rcCtxModel.getState()][0];
#else
		UInt logprobLPS = TComPACoderTables::ilogprobMPS2logprobLPS[ logprobMPS-1 ];
#endif

		if( logprbOfVal > (logprobLPS+logprob) )//fast channel for MPS
		{
			logprob += logprobLPS;
			ruiBin      = 0;
			rcCtxModel.updateLPS();
		}
		else
		{
			UInt thrshld = dvsn_intervalNoRenorm(logprobLPS+logprob);
			if(thrshld <= value)
			{
				value -= thrshld;
				logprob += logprobMPS;

				ruiBin      = 1;
				rcCtxModel.updateMPS();
				mapVal2LngPrbNoRenorm(value, logprob, logprbOfVal)
			}
			else
			{
				logprob += logprobLPS;
				ruiBin      = 0;
				rcCtxModel.updateLPS();
			}
		}
	}
	
	if(logprob >= 0x800)
	{
		logprob -= 0x800;
		
		m_pcTComBitstream->read(16);
		iLastDWORD = m_pcTComBitstream->peekBits(32);
		value = (value<<16) + (iLastDWORD&0xffff);

		mapVal2LngPrbNoRenorm(value, logprob, logprbOfVal)
	}
}
#endif

Void
TDecBinPACoder::decodeBinEP( UInt& ruiBin )
{
	logprob += 128;
	UInt thrshld = dvsn_intervalNoRenorm(logprob);

	ruiBin = 0;
	if(thrshld <= value)
	{
		value -= thrshld;
		ruiBin = 1;

		mapVal2LngPrbNoRenorm(value, logprob, logprbOfVal)
	}

	if(logprob >= 0x800)
	{
		logprob -= 0x800;
		/*xReadByte(value);
	  xReadByte(value);*/
		m_pcTComBitstream->read(16);
		iLastDWORD = m_pcTComBitstream->peekBits(32);
		value = (value<<16) + (iLastDWORD&0xffff);

		mapVal2LngPrbNoRenorm(value, logprob, logprbOfVal)
	}
}


Void
TDecBinPACoder::decodeBinTrm( UInt& ruiBin )
{
	logprob += 1;
	UInt thrshld = dvsn_intervalNoRenorm(logprob);
//print_state();
  if(thrshld <= value)
  {
    ruiBin = 1;
  }
  else
  {
    ruiBin = 0;
		if(logprob >= 0x800)
		{
			logprob -= 0x800;
			/*xReadByte(value);
			xReadByte(value);*/
			m_pcTComBitstream->read(16);
			iLastDWORD = m_pcTComBitstream->peekBits(32);
			value = (value<<16) + (iLastDWORD&0xffff);

			mapVal2LngPrbNoRenorm(value, logprob, logprbOfVal)
		}
  }
}


Void  
TDecBinPACoder::xReadByte( UInt& ruiVal )
{
  UInt uiBit = 0;
  m_pcTComBitstream->read(8, uiBit );
  ruiVal  = ( ruiVal << 8 ) | uiBit;
}

#if E057_INTRA_PCM
/** Reset BAC register values.
 * \returns Void
 */
Void TDecBinPACoder::resetBac()
{
  logprob = 0;
	value = iLastDWORD = m_pcTComBitstream->peekBits(32);
  /*value   = 0;
	xReadByte(value);
	xReadByte(value);
	xReadByte(value);
	xReadByte(value);*/

	mapVal2LngPrbNoRenorm(value, logprob, logprbOfVal)
}
/** Decode PCM alignment zero bits.
 * \returns Void
 */
Void TDecBinPACoder::decodePCMAlignBits()
{
	Int ii;
  UInt uprng, valueBttm;
	valueBttm = iLastDWORD - value;
	uprng = valueBttm + dvsn_intervalNoRenorm(logprob);
//printf("pcm mode!\n");
	if(uprng < valueBttm)
	{
		for(ii=31; ii>=0; ii--)
		{
			UInt valcur = (valueBttm|(1<<ii)) & (~((1<<ii)-1));
			if( valcur >= valueBttm)
			{
				break;
			}
		}
	}
	else
	{
		for(ii=31; ii>=0; ii--)
		{
			UInt valcur = (valueBttm|(1<<ii)) & (~((1<<ii)-1));
			UInt addvalff = (1<<(ii&0x18)) - 1;
			
			if( (valcur>=valueBttm) && (valcur+addvalff<uprng) )
			{
				break;
			}
		}
	}

	Int iPassedBytes=0;
	while(ii<32)
	{
		iPassedBytes++;
		ii += 8;
	}
	m_pcTComBitstream->read(iPassedBytes*8);
}

/** Read a PCM code.
 * \param uiLength code bit-depth
 * \param ruiCode pointer to PCM code value
 * \returns Void
 */
Void  TDecBinPACoder::xReadPCMCode(UInt uiLength, UInt& ruiCode)
{
  assert ( uiLength > 0 );
  m_pcTComBitstream->read (uiLength, ruiCode);
}
#endif
