/* The copyright in this software is being made available under the BSD
 * License, included below. This software may be subject to other third party
 * and contributor rights, including patent rights, and no such rights are
 * granted under this license.
 *
 * Copyright (c) 2010-2011, ITU/ISO/IEC
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *  * Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *  * Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *  * Neither the name of the ITU/ISO/IEC nor the names of its contributors may
 *    be used to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 */

/** \file     TEncBinPACoder.cpp
    \brief    binary entropy encoder of PA-Coder
*/

#include "TEncBinPACoder.h"


TEncBinPACoder::TEncBinPACoder()
: m_pcTComBitIf( 0 )
{
}


TEncBinPACoder::~TEncBinPACoder()
{
}


Void
TEncBinPACoder::init( TComBitIf* pcTComBitIf )
{
  m_pcTComBitIf = pcTComBitIf;
}


Void
TEncBinPACoder::uninit()
{
  m_pcTComBitIf = 0;
}


Void
TEncBinPACoder::start()
{
	low = 0;
	logprob = 0;
	numOfBytes255 = value = backupbyte = 0;
	endhigh = 0;
}


Void
TEncBinPACoder::finish()
{
	encode_btrmend();
}

#if E057_INTRA_PCM
/** Reset BAC register and counter values.
 * \returns Void
 */
Void TEncBinPACoder::resetBac()
{
  low = 0;
	logprob = 0;
	numOfBytes255 = value = backupbyte = 0;
	endhigh = 0;
}

/** Encode PCM alignment zero bits.
 * \returns Void
 */
Void TEncBinPACoder::encodePCMAlignBits()
{
	//endhigh = low + dvsn_intervalNoRenorm(logprob);
  encode_btrmend();
}

/** Write a PCM code.
 * \param uiCode code value
 * \param uiLength code bit-depth
 * \returns Void
 */
Void  TEncBinPACoder::xWritePCMCode(UInt uiCode, UInt uiLength)
{
  m_pcTComBitIf->write  (uiCode, uiLength);
}
#endif


Void
TEncBinPACoder::copyState( TEncBinIf* pcTEncBinIf )
{
  TEncBinPACoder* pcTEncBinPACoder = pcTEncBinIf->getTEncBinPACoder();

	low           = pcTEncBinPACoder->low; 
	logprob       = pcTEncBinPACoder->logprob;
	endhigh       = pcTEncBinPACoder->endhigh;
  value         = pcTEncBinPACoder->value;
	numOfBytes255 = pcTEncBinPACoder->numOfBytes255;
	backupbyte    = pcTEncBinPACoder->backupbyte;
}

Void  
TEncBinPACoder::resetBits()
{
	logprob      &= 0x7f;
	numOfBytes255 = 0; 
	value         = 0;
	low           = 0;
}


UInt
TEncBinPACoder::getNumWrittenBits()
{
  return m_pcTComBitIf->getNumberOfWrittenBits() + 8*numOfBytes255 + (logprob>>7);
}


Void
TEncBinPACoder::encodeBin( UInt uiBin, ContextModel &rcCtxModel )
{
#if HHI_RQT
  {
    DTRACE_CABAC_V( g_nSymbolCounter++ )
    DTRACE_CABAC_T( "\tstate=" )
    DTRACE_CABAC_V( ( rcCtxModel.getState() << 1 ) + rcCtxModel.getMps() )
    DTRACE_CABAC_T( "\tsymbol=" )
    DTRACE_CABAC_V( uiBin )
    DTRACE_CABAC_T( "\n" )
  }
#endif
	if( uiBin != rcCtxModel.getMps() )
  {
#ifndef PACoder_Prob_Est
		UInt logprblow = logprob + TComPACoderTables::lgprobMLpsNoRenorm[rcCtxModel.getState()][1];
		logprob += TComPACoderTables::lgprobMLpsNoRenorm[rcCtxModel.getState()][0];
#else
		UInt logprblow = logprob + rcCtxModel.getState() + 1;
		logprob += TComPACoderTables::ilogprobMPS2logprobLPS[ rcCtxModel.getState() ];
#endif

		low += dvsn_intervalNoRenorm(logprblow);
		rcCtxModel.updateLPS();
  }
  else
  {
#ifndef PACoder_Prob_Est
		logprob += TComPACoderTables::lgprobMLpsNoRenorm[rcCtxModel.getState()][1];
#else
		logprob += (rcCtxModel.getState() + 1);
#endif
		rcCtxModel.updateMPS();
  }

	if(logprob >= 0x800)
	{
		logprob -= 0x800;
		encode_bitstream();
	}
}

#ifdef PACoder_LowBound0
Void  TEncBinPACoder::encodeBinL0( UInt  uiBin,  ContextModel& rcCtxModel )
{
	if(rcCtxModel.getMps()==0)
	{
		if( uiBin != 0 )
		{
#ifndef PACoder_Prob_Est
			UInt logprblow = logprob + TComPACoderTables::lgprobMLpsNoRenorm[rcCtxModel.getState()][1];
			logprob += TComPACoderTables::lgprobMLpsNoRenorm[rcCtxModel.getState()][0];
#else
			UInt logprblow = logprob + rcCtxModel.getState() + 1;
			logprob += TComPACoderTables::ilogprobMPS2logprobLPS[ rcCtxModel.getState() ];
#endif

			low += dvsn_intervalNoRenorm(logprblow);
			rcCtxModel.updateLPS();
		}
		else
		{
#ifndef PACoder_Prob_Est
			logprob += TComPACoderTables::lgprobMLpsNoRenorm[rcCtxModel.getState()][1];
#else
			logprob += (rcCtxModel.getState() + 1);
#endif
			rcCtxModel.updateMPS();
		}
	}
	else
	{
		if( uiBin != 0 )
		{
#ifndef PACoder_Prob_Est
			UInt logprblow = logprob + TComPACoderTables::lgprobMLpsNoRenorm[rcCtxModel.getState()][0];
			logprob += TComPACoderTables::lgprobMLpsNoRenorm[rcCtxModel.getState()][1];
#else
			UInt logprblow = logprob + TComPACoderTables::ilogprobMPS2logprobLPS[ rcCtxModel.getState() ];
			logprob += rcCtxModel.getState() + 1;
#endif

			low += dvsn_intervalNoRenorm(logprblow);
			rcCtxModel.updateMPS();
		}
		else
		{
#ifndef PACoder_Prob_Est
			logprob += TComPACoderTables::lgprobMLpsNoRenorm[rcCtxModel.getState()][0];
#else
			logprob += TComPACoderTables::ilogprobMPS2logprobLPS[ rcCtxModel.getState() ];
#endif
			rcCtxModel.updateLPS();
		}
	}

	if(logprob >= 0x800)
	{
		logprob -= 0x800;
		encode_bitstream();
	}
}
#endif

Void
TEncBinPACoder::encodeBinEP( UInt uiBin )
{
#if HHI_RQT
  {
    DTRACE_CABAC_V( g_nSymbolCounter++ )
    DTRACE_CABAC_T( "\tEPsymbol=" )
    DTRACE_CABAC_V( uiBin )
    DTRACE_CABAC_T( "\n" )
  }
#endif
	logprob += 128;
	if( uiBin )
  {
		low += dvsn_intervalNoRenorm(logprob);
  }

	if(logprob >= 0x800)
	{
		logprob -= 0x800;
		encode_bitstream();
	}
}


Void
TEncBinPACoder::encodeBinTrm( UInt uiBin )
{
	if(uiBin == 0)
  {
	  logprob += 1;

	  if(logprob >= 0x800)
		{
			logprob -= 0x800;
			encode_bitstream();
		}
  }
  else
  {
	  endhigh = low + dvsn_intervalNoRenorm(logprob);
	  low += dvsn_intervalNoRenorm(logprob+1);
  }
}

Void TEncBinPACoder::encode_bitstream()
{
	UInt isbigthan32bit = ((0xffffffff-value)<low);
	UInt value2bytes;
//printf("%10x\n", dvsn_intervalNoRenorm(0xd9));
	value += low;
	low = 0;
	value2bytes = value>>16;
	value <<= 16;
//printZ("value %10x  logprob %4x  bckbyte %2x  numbytes255 %4d value2bytes %6x\n", value, logprob, backupbyte,numOfBytes255,value2bytes);
	if(isbigthan32bit)
	{
		m_pcTComBitIf->write( backupbyte+1, 8 );
		numOfBytes255--;
		if(numOfBytes255)
		{
			do
			{
				m_pcTComBitIf->write( 0, 8 );
				numOfBytes255--;
			}
			while(numOfBytes255);
		}
	}
	
	if(value2bytes == 0xffff)
	{
		if(numOfBytes255)
			numOfBytes255 += 2;
		else
			m_pcTComBitIf->write( 0xffff, 16 );
	}
	else
	{
		if(numOfBytes255)
		{
			m_pcTComBitIf->write( backupbyte, 8 );
			numOfBytes255--;
			if(numOfBytes255)
			{
				do
				{
					m_pcTComBitIf->write( 0xff, 8 );
					numOfBytes255--;
				}
				while(numOfBytes255);
			}
		}

		if( (value2bytes&0xff) == 0xff )
		{
			backupbyte = value2bytes>>8;
			numOfBytes255 = 2;
		}
		else
		{
			m_pcTComBitIf->write( value2bytes>>8, 8 );
			backupbyte = value2bytes&0xff;
			numOfBytes255 = 1;
		}
	}
//printZ("value %10x  logprob %4x  bckbyte %2x  numbytes255 %4d\n", value, logprob, backupbyte,numOfBytes255);
}
Void TEncBinPACoder::encode_btrmend()
{
	UInt isbigthan32bit = ((0xffffffff-value)<low);
	UInt uprng = endhigh+value, valuewrite;
	Int ii, endbit;
	value += low;
//printZ("value %10x  uprng %10x isbigthan32bit %d   numOfBytes255 %d\n", value, uprng, isbigthan32bit, numOfBytes255);
	if(isbigthan32bit)
	{
		m_pcTComBitIf->write( backupbyte+1, 8 );
		numOfBytes255--;
		if(numOfBytes255)
		{
			do
			{
				m_pcTComBitIf->write( 0, 8 );
				numOfBytes255--;
			}
			while(numOfBytes255);
		}
	}
	
	if(numOfBytes255)
	{
		m_pcTComBitIf->write( backupbyte, 8 );
		numOfBytes255--;
		if(numOfBytes255)
		{
			do
			{
				m_pcTComBitIf->write( 0xff, 8 );
				numOfBytes255--;
			}
			while(numOfBytes255);
		}
	}
	if(uprng < value)
	{
		for(ii=31; ii>=0; ii--)
		{
			UInt valcur = (value|(1<<ii)) & (~((1<<ii)-1));
			if( valcur >= value)
			{
				valuewrite = valcur;
				break;
			}
		}
	}
	else
	{
		for(ii=31; ii>=0; ii--)
		{
			UInt valcur = (value|(1<<ii)) & (~((1<<ii)-1));
			UInt addvalff = (1<<(ii&0x18)) - 1;
			
			if( (valcur>=value) && (valcur+addvalff<uprng) )
			{
				valuewrite = valcur;
				break;
			}
		}
	}
//printZ("valuewrite %10x num %d\n", valuewrite, (32-ii+7)>>3);
	while(ii<32)
	{
		m_pcTComBitIf->write( (valuewrite>>24), 8 );
		valuewrite <<= 8;
		ii += 8;
	}
}
